from inverse import inv
from glob import globs
from vbweights import w
from counters import counter
from assignment import mrv_variables, initialize_mrv_vars, add_var, remove_var
from collections import deque
from learned_weights import learnedWeights
from helpfuncs import bitdecoding

if globs['split'] == 0: # splitting based on set of base relations
   from basesplit import bsplit as split
elif globs['split'] == 1: # splitting based on horn set
   from hornsplit import hsplit as split

if globs['pcheuristic'] == 0: # path consistency
   from pc import PC as pc
elif globs['pcheuristic'] == 1: # path consistency with weighted queue
   from pcw import PCw as pc

if globs['heuristic'] == 0: # none
   def myvar((i,j), neighbors, ConMatrix):
      for ci,cj in mrv_variables:
         dummy, (i,j) = 0, (ci,cj)
         return dummy, (i,j)
      return False
elif globs['heuristic'] == 1: # local
   def myvar((i,j), neighbors, ConMatrix):        
      clique = []
      for k in neighbors[i] & neighbors[j]:
         clique.append(tuple(sorted((i,k))))
         clique.append(tuple(sorted((j,k))))
      candidates = [(i,j) for i,j in clique if (i,j) in mrv_variables]
      if candidates:
         selection = [(i,j) for i,j in learnedWeights if (i,j) in candidates]
         if not selection:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1]),(i,j)) for i,j in candidates])
         else:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1])/float(learnedWeights[(i,j)]),(i,j)) for i,j in selection])
         return dummy, (i,j) 
      elif mrv_variables:
         selection = [(i,j) for i,j in learnedWeights if (i,j) in mrv_variables]
         if not selection:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1]),(i,j)) for i,j in mrv_variables])
         else:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1])/float(learnedWeights[(i,j)]),(i,j)) for i,j in selection])
         return dummy, (i,j)
      return False
elif globs['heuristic'] == 2: # global
   def myvar((i,j), neighbors, ConMatrix):
      try:
         selection = [(i,j) for i,j in learnedWeights if (i,j) in mrv_variables]
         if not selection:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1]),(i,j)) for i,j in mrv_variables])
         else:
            dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1])/float(learnedWeights[(i,j)]),(i,j)) for i,j in selection])
      except:
         return False
      return dummy, (i,j)

cardmult = max(w)
lastAssigned = None

def nbUnsatisfiedConstraints(ConMatrix, IConMatrix):

    count = 0

    for i in xrange(len(ConMatrix)):
       for j in xrange(i+1,len(ConMatrix)):
           if not ConMatrix[i][j] & IConMatrix[i][j]:
              count += 1

    return count

# backtracking consistency
def consistency(ConMatrix, IConMatrix, neighbors, alpha, k, splittings, m = -1, n = -1, repairs = None, trail =  None):  

   counter['nodes']+=1 # increment visited nodes counter
   global lastAssigned

   if m == -1: 
      initialize_mrv_vars(ConMatrix, IConMatrix, neighbors)
      learnedWeights.clear()
   else:
      if not pc(ConMatrix, IConMatrix, neighbors, m, n, repairs, trail):
         return None, alpha

   #print alpha
   if nbUnsatisfiedConstraints(ConMatrix, IConMatrix) >= alpha - k:
      return None, alpha 

   counter['decisions'] += 1

   if not lastAssigned:
      try:
         dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1]),(i,j)) for i,j in mrv_variables])
         lastAssigned = (i,j)
      except:
         alpha = nbUnsatisfiedConstraints(ConMatrix, IConMatrix)
         return ConMatrix, alpha
   else:
      myNextVar = myvar(lastAssigned, neighbors, ConMatrix)
      if myNextVar:
         dummy, (i,j) = myNextVar
         lastAssigned = (i,j)
      else:
         alpha = nbUnsatisfiedConstraints(ConMatrix, IConMatrix)
         return ConMatrix, alpha

   mysplits = len(bitdecoding(ConMatrix[i][j]))
   # splitting
   try:
      mytemp = reduce(lambda x, y: x | y, [z for z in set(bitdecoding(ConMatrix[i][j])) - set(bitdecoding(IConMatrix[i][j]))])
   except:
      mytemp = None   
   myrel = set(split[mytemp-1][1][:]) if mytemp else set()
   try:
      mytemp = reduce(lambda x, y: x | y, [z for z in set(bitdecoding(IConMatrix[i][j])) & set(bitdecoding(ConMatrix[i][j]))])
   except:
      mytemp = None
   myrelprime = set(split[mytemp-1][1][:]) if mytemp else set()

   d = list(myrel | myrelprime)

   splittings.append(len(d)/float(mysplits))
   
   # least constraining value heuristic
   dw = [((-w[a-1] if a not in myrelprime else -w[a-1]*cardmult),a) for a in d]
   dw.sort()
   dw = [a[1] for a in dw]

   repairs = {}
   trail = deque([])

   result = None

   # as long as a consistent variable-value pair in not found, search for it
   for v in dw:
       repairs[(i,j)] = ConMatrix[i][j]
       # assignment takes place
       ConMatrix[i][j] = v
       ConMatrix[j][i] = inv[v-1]
       remove_var((i,j))
       scenario, alpha = consistency(ConMatrix, IConMatrix, neighbors, alpha, k, splittings, i, j, repairs, trail) # keep copy of the constraint matrix in case an inconsistency happens
       if (scenario != None):
          result = tuple([ic[:] for ic in scenario])
          alpha = nbUnsatisfiedConstraints(scenario, IConMatrix)
       # revert contraint mantrix and mrv variables to previous state
       for g, h in repairs.keys():
          ConMatrix[g][h] = repairs[(g, h)] 
          ConMatrix[h][g] = inv[repairs[(g, h)]-1]
       add_var((i,j))
       while trail:
          c,(a,b) = trail.pop()
          remove_var((a,b)) if c else add_var((a,b))            

   return result, alpha
